"""
Health checks - simple status reporters for monitoring / status pages
"""

import logging
from abc import ABC, abstractmethod, abstractproperty
import datetime

from flask import Markup

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


MONDAY = 0
TUESDAY = 1
WEDNESDAY = 2
THURSDAY = 3
FRIDAY = 4
SATURDAY = 5
SUNDAY = 6


class HealthCheckException(Exception):
    pass


class HealthCheckError(HealthCheckException):
    pass


class HealthCheckWarning(HealthCheckException):
    pass


class _HealthCheckSchedule(ABC):
    @abstractmethod
    def should_run(self, dt):
        """
        :param dt: Datetime
        :return: Whether or not the healthcheck should run
        """
        pass


class AlwaysRun(_HealthCheckSchedule):
    def should_run(self, dt):
        return True


ALWAYS_RUN = AlwaysRun()


class IncludeDaysOfWeek(_HealthCheckSchedule):
    def __init__(self, days_of_week):
        """
        :param days_of_week: List of integers with day of week to
                             run the health checks (0 is Monday, 6 is Sunday)
        """
        self.days_of_week = days_of_week

    def should_run(self, dt):
        return dt.date().weekday in self.days_of_week


class ExcludeDaysOfWeek(_HealthCheckSchedule):
    def __init__(self, days_of_week):
        """
        :param days_of_week: List of integers with day of week to NOT
                             run the health checks (0 is Monday, 6 is Sunday)
        """
        self.days_of_week = days_of_week

    def should_run(self, dt):
        return dt.date().weekday() not in self.days_of_week


class DayInterval:
    def __init__(self, from_time, to_time, include=True):
        self.from_time = from_time
        self.to_time = to_time
        self.include = include

    def should_run(self, dt):
        time = dt.time()

        if self.from_time is not None and time < self.from_time:
            return not self.include

        if self.to_time is not None and time > self.to_time:
            return not self.include

        return self.include

    def __repr__(self):
        return '{} {} to {}'.format(
            'Include' if self.include else 'Exclude',
            self.from_time if self.from_time is not None else '0:00',
            self.to_time if self.to_time is not None else '24:00'
        )


class IncludeInterval(_HealthCheckSchedule):
    def __init__(self, from_day, from_time, to_day, to_time):
        if from_day < 0:
            raise ValueError('From day must be >= 0')
        if from_day > 6:
            raise ValueError('From day must be <= 6')
        if to_day < 0:
            raise ValueError('To day must be >= 0')
        if to_day > 6:
            raise ValueError('To day must be <= 6')

        self.from_day = from_day
        self.from_time = from_time
        self.to_day = to_day
        self.to_time = to_time

        self.daily_schedules = [None] * 7

        if from_day == to_day:
            if from_time == to_time:
                raise ValueError('From and to times are the same')
            elif from_time < to_time:
                # This only runs on a specific day
                self.daily_schedules[from_day] = DayInterval(from_time, to_time)
            else:
                # This only excludes a specific day
                for i in range(0, 7):
                    if i != from_day:
                        self.daily_schedules[i] = DayInterval(None, None)

                self.daily_schedules[from_day] = DayInterval(from_time, to_time, include=False)
        elif from_day < to_day:
            for i in range(0, 7):
                if i < from_day:
                    pass
                elif i == from_day:
                    self.daily_schedules[i] = DayInterval(from_time, None)
                elif i < to_day:
                    self.daily_schedules[i] = DayInterval(None, None)
                elif i == to_day:
                    self.daily_schedules[i] = DayInterval(None, to_time)
                # else pass
        else:
            assert to_day < from_day
            for i in range(0, 7):
                if i < to_day:
                    self.daily_schedules[i] = DayInterval(None, None)
                elif i == to_day:
                    self.daily_schedules[i] = DayInterval(None, to_time)
                elif i < from_day:
                    pass
                elif i == from_day:
                    self.daily_schedules[i] = DayInterval(from_time, None)
                else:
                    self.daily_schedules[i] = DayInterval(None, None)

        # print(self.daily_schedules)
    
    def should_run(self, dt):
        weekday = dt.date().weekday()
        daily_schedule = self.daily_schedules[weekday]
        
        if daily_schedule is None:
            return False
        
        return daily_schedule.should_run(dt)


class ExcludeInterval(IncludeInterval):
    def should_run(self, dt):
        return not super().should_run(dt)


class HealthCheck(ABC):
    @abstractproperty
    def name(self):
        pass

    @abstractmethod
    def check(self):
        """
        :return a success message (an exception will be raised if this fails)
        """
        pass
    
    @property
    def schedule(self):
        return ALWAYS_RUN

    def should_run(self, dt=None):
        if dt is None:
            dt = datetime.datetime.utcnow()

        return self.schedule.should_run(dt)


class HealthCheckResult:
    def __init__(self, name, status, message):
        self.name = name
        self.status = status
        self.message = message


def run_health_checks(health_checks):
    results = []
    
    for health_check in health_checks:
        if health_check.should_run():
            status = 'success'
            try:
                message = health_check.check()
            except HealthCheckError as e:
                message = str(e)
                status = 'error'
            except HealthCheckWarning as w:
                message = str(w)
                status = 'warning'

            results.append(HealthCheckResult(health_check.name, status, message))

    return results


def render_health_check_results(results):
    out = ['<table class="health-checks"><tbody>']
    for result in results:
        out.append(
            '<tr><th>{}</th><td class="{}">{}</td></tr>'.format(
                result.name, result.status, result.message
            )
        )
    out.append('</tbody></table>')
    
    return Markup(''.join(out))


def run_and_render_health_checks(health_checks=None):
    results = run_health_checks(health_checks)

    return render_health_check_results(results)



"""
System for loading YAML Files into objects (i.e. database models)
"""

import logging
import os
import decimal
import re

import yaml
from littlefish import util

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

_HTML_COLOUR_REGEX = re.compile(r'^#[0-9a-f]{6}$')


class DataType(object):
    def __init__(self, name, from_yaml, constant_value=None):
        """
        :param name: The name of the type
        :param from_yaml: Function to convert from yaml variable to Python datatype
        :param constant_value: If set, this will always have the same value and will
                               not be loaded from the YAML data
        """
        self.name = name
        self.from_yaml = from_yaml
        self.constant_value = constant_value


# This is mainly used for testing - returns whatever the yaml datatype happens to be
RAW = DataType('!RAW!', lambda x: x)

# This is a special case - fields of this will not be parsed at all, will not be passed into the
# constructor and will not be updated. Use this for sub-models
IGNORE = DataType('!IGNORE!', None)


# This is another special case - use this when some extra field is needed for the import, that
# isn't in the yaml.  This is used when a relation needs to be imported and the original object
# that it's attached to needs to be passed through
def CONSTANT(value):  # noqa
    if value is None:
        raise ValueError('Constant value can\'t be None!')

    return DataType('!CONSTANT!', None, constant_value=value)


def load_single_line_str(s):
    """validates that a string is single line"""
    s = str(s)

    if '\n' in s:
        raise ValueError('String value "{}" contains newline character. HINT: consider using '
                         'yaml ">-" operator for multi-line strings'.format(s))

    return s


# STR only allows single line strings
STR = DataType('str', load_single_line_str)

# MULTI_LINE_STR allows any string with no validation
MULTILINE_STR = DataType('str/multi-line', lambda s: str(s))


def load_int(i):
    if isinstance(i, float):
        raise ValueError('Float value {} passed when integer required'.format(i))
    
    return int(i)


INT = DataType('int', load_int)


def load_decimal(s):
    if isinstance(s, float):
        raise ValueError('Float value {} passed when Decimal required. Please use quoted '
                         'string instead'.format(s))

    return decimal.Decimal(s)


DECIMAL = DataType('Decimal', load_decimal)


def load_html_colour(s):
    s = s.lower().strip()

    if not _HTML_COLOUR_REGEX.match(s):
        raise ValueError('Invalid HTML colour: "{}"'.format(s))

    return s


HTML_COLOUR = DataType('html_colour', load_html_colour)


def load_boolean(b):
    if not isinstance(b, bool):
        raise ValueError('Boolean required, got "{}"'.format(b))

    return b


BOOLEAN = DataType('bool', load_boolean)


class ImportField(object):
    def __init__(self, name, data_type, constructor_field=True, identifier=False,
                 optional=False, default=None, unchangeable=False, no_update=False,
                 python_name=None):
        """
        :param name: The field in the yaml file. By default this is also the spelling of the
                     python field and keyword argument for constructor
        :param data_type: A DataType defining the type of this field
        :param constructor_field: Is this field passed in to the constructor
        :param identifier: If True, this field cannot be changed once it is set and is used to
                           load the object from the database to update it (i.e. it is a unique
                           identifier of the object)
        :param optional: If True, this field doesn't need to be present to import the object. It
                         will be set to None if not present
        :param default: Default value for optional arguments that are not present
        :param unchangeable: If True, this can never be updated after it has been inserted
        :param no_update: This will only be set for new models - existing models will not be updated.
                          The difference between this and unchangeable is that unchangeable fields will
                          raise an error if the value in the yaml file is different from the database,
                          but these will not.
        :param python_name: Use this to override the spelling of the python variable in the model
        """
        if default is not None and not optional:
            raise ValueError('Only optional arguments can have a default!')

        self.name = name
        self.data_type = data_type
        self.constructor_field = constructor_field
        self.identifier = identifier
        self.optional = optional
        self.default = default
        self.unchangeable = unchangeable
        self.no_update = no_update
        self._python_name = python_name

    def from_yaml(self, s):
        try:
            return self.data_type.from_yaml(s)
        except Exception as e:
            error = str(e)
            raise ValueError('Error reading field "{}": {}'.format(self.name, error))

    @property
    def constant_value(self):
        return self.data_type.constant_value

    @property
    def python_name(self):
        if self._python_name:
            return self._python_name

        return self.name


def get_list(data_dict, key, optional=False):
    """
    Extract a list from a data_dict and do a bit of error checking
    """
    if key not in data_dict:
        if optional:
            return None
        else:
            raise ValueError('Key "{}" not present in data'.format(key))

    l = data_dict[key]  # noqa
    if not isinstance(l, list):
        raise ValueError('Key "{}" is not a list'.format(key))

    return l


def load_yaml_from_directory(path):
    """
    Load yaml files from a directory and return a generator with the parsed dicts

    :return: Generator of tuples (filename, data_dict)
    """
    files = os.listdir(path)
    
    for f in files:
        filename = os.path.join(path, f)
        if os.path.isfile(filename) and not f.startswith('.'):
            log.info('Loading {}'.format(filename))
            with open(filename, 'r') as stream:
                data = yaml.load(stream)

                yield filename, data


def import_model(model_class, yaml_data_dict, fields, session, commit=False, silence_no_changes=False,
                 insert_only=False):
    """
    Import a model from a parsed yaml file

    :param model_class: Model class, i.e. Category
    :param data_dict: Dictionary obtained by parsing the yaml file
    :param fields: List of ImportFields defining the structure of the yaml file
    :param session: Database session (i.e. db.session)
    :param commit: Whether to commit to the database (defaults to False)
    :param silence_no_changes: Don't emit a log message when there are no changes
    :param insert_only: Don't update existing data - only insert new data
    """
    model_name = model_class.__name__

    # Make sure there are no duplicates in the fields
    all_field_yaml_names = set()
    all_field_python_names = set()
    for field in fields:
        if field.name in all_field_yaml_names:
            raise ValueError('Multiple fields in list with same name "{}"'.format(field.name))
        all_field_yaml_names.add(field.name)
        
        if field.python_name in all_field_python_names:
            raise ValueError('Miltiple fields in list with same python name "{}"'.format(field.python_name))
        all_field_python_names.add(field.python_name)
    
    # This may seem a little backwards - we process the data first and then validate it afterwards
    # Create a new dict with a combination of the data from data_dict and any relevant defaults.
    # This dict maps name -> value
    data = {}
    for field in fields:
        if field.data_type == IGNORE:
            # Do nothing with this field
            continue

        if field.constant_value:
            data[field.name] = field.constant_value
        elif field.name in yaml_data_dict:
            raw_value = yaml_data_dict[field.name]
            if raw_value is None:
                data[field.name] = None
            else:
                data[field.name] = field.from_yaml(raw_value)
        elif field.optional:
            data[field.name] = field.default

    # Data now contains processed versions of all of the fields in the yaml data

    # Next let's find the identifiers - the fields used to deduplicate this model and to update
    # existing models instead of creating new ones.  We will store these in a dict mapping yaml
    # name -> python name
    identifier_keys = {}
    for field in fields:
        if field.identifier:
            identifier_keys[field.name] = field.python_name
    
    # Loop over python names
    for identifier_key in identifier_keys.values():
        if identifier_key not in data:
            raise ValueError('Indentifier field "{}" not present in data when loading object of '
                             'type {}'.format(identifier_key, model_name))

    # Get the identifier for logging
    identifier = '/'.join([str(data[i]) for i in identifier_keys.values()])
    
    # These lists use the yaml_names
    all_field_keys = []
    required_field_keys = []
    
    # Now that we've loaded the data - it's time to validate that the correct stuff is present!
    for field in fields:
        if not field.constant_value:
            all_field_keys.append(field.name)
            if not field.optional:
                required_field_keys.append(field.name)
    
    # Check that the correct fields have been supplied
    for field in required_field_keys:
        if field not in yaml_data_dict:
            raise Exception('Required field "{}" not present in {} "{}"'.format(
                field, model_name, identifier
            ))

    for field in yaml_data_dict:
        if field not in all_field_keys:
            raise Exception('Unknown field "{}" found in {} "{}"'.format(
                field, model_name, identifier
            ))

    # Create a list of SQLAlchemy filters to load the model if it exists already
    filters = []
    for (yaml_key, python_name) in identifier_keys.items():
        filters.append(getattr(model_class, python_name) == data[yaml_key])
    
    # Load the instance to see if it exists
    model_instance = model_class.query.filter(*filters).one_or_none()
    updating = model_instance is not None

    if updating and insert_only:
        if not silence_no_changes:
            log.info(' X  Not Updating {} "{}"'.format(model_name, identifier))

        return model_instance

    # Convert values and assign to dicts to either pass into constructor, or update the field.
    # We start using python_name instead of name from here on as these dicts will be passed into
    # constructors etc.
    kwargs = {}
    update_fields = {}
    unchangeable_fields = set()

    for field in fields:
        if field.data_type == IGNORE:
            # Don't process this field
            continue
        
        if field.unchangeable:
            unchangeable_fields.add(field.python_name)
        
        # First extract the field from the dict and convert
        if field.name in data:
            # Field is present in parsed file
            value = data[field.name]
        else:
            # This should be impossible
            raise Exception('Required field not present in data dic')

        if not model_instance and field.constructor_field:
            kwargs[field.python_name] = value
        elif not field.identifier and not field.no_update:
            update_fields[field.python_name] = value

    # Now we have our field dicts - time to insert / update the object
    if not model_instance:
        log.info(' * Adding {} "{}"'.format(model_name, identifier))
        model_instance = model_class(**kwargs)
        session.add(model_instance)

    update_logger = util.UpdateLogger('   * ', use_logger=True)

    for field in update_fields:
        new_val = update_fields[field]
        if updating:
            old_val = getattr(model_instance, field)
            if old_val != new_val:
                # Check for unchangeable field being updated
                if field in unchangeable_fields:
                    raise Exception(
                        'Trying to change unchangeable field from {} to {}'.format(
                            old_val, new_val
                        )
                    )

                if isinstance(old_val, str) and len(old_val) > 100 \
                        or isinstance(new_val, str) and len(new_val) > 100:
                    update_logger.log_variable_update(field)
                else:
                    update_logger.log_variable_update(field, old_val, new_val)
        setattr(model_instance, field, new_val)

    if update_logger.has_updates:
        log.info(' * Updating {} "{}"'.format(model_name, identifier))
        update_logger.print_updates()
    elif updating and not silence_no_changes:
        log.info(' * (no changes for {} "{}")'.format(model_name, identifier))
    
    if commit:
        session.commit()

    return model_instance




"""
Base type for types and statuses - python classes for storing types / statuses, ensuring that they
render as a span with a css class and tooltip, and adding database types to automatically
decode them from strings in the database. I have elected not to use python enums because I
hate the way that they handle properties!
"""

import logging

import sqlalchemy.types

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class MetaEnumeration(type):
    def __call__(cls, *args, **kwargs):
        # Create and initialise the instance
        instance = super().__call__(*args, **kwargs)
        
        if not hasattr(cls, 'instances'):
            cls.instances = {}
        
        if instance.code in cls.instances:
            raise ValueError('Code must unique. An instance of {} already exists '
                             'with code {}'.format(cls.__name__, instance.code))

        cls.instances[instance.code] = instance

        return instance


class Enumeration(metaclass=MetaEnumeration):
    """
    Subclass this to create classes for different "Types" and "Statuses", i.e.
    OrderStatus, StaffActionType.

    These are basically enumerations.  They are nowehere near as safe as the standard library
    enum but are more convenient for adding attributes.

    You can use Enumeration.instances to get a dict of instances mapping code onto the instance,
    Enumeration.from_code(code) to get an instance from a code and Enumeration.db_type() to
    get the SQLAlchemy column type.  In jinja scripts you can output them to get a nice span
    with a css class and a tooltip.

    Be sure to override the "css_class" class variable for html rendering.
    """
    # Set this in subclass to change the base css class
    base_css_class = 'unknown'

    def __init__(self, code, description):
        self.code = code
        self.description = description

    @property
    def css_class(self):
        return '{} {}-{}'.format(
            self.base_css_class, self.base_css_class, self.code.replace('_', '-').replace(' ', '-').lower()
        )

    def __html__(self):
        return '<span data-toggle="tooltip" title="{}" class="{}">{}</span>'.format(
            self.description, self.css_class, self.code
        )

    def __str__(self):
        return self.code

    @classmethod
    def from_code(cls, code):
        if not hasattr(cls, 'instances') or code not in cls.instances:
            raise ValueError('Code "{}" is not a valid {}'.format(
                code, cls.__name__
            ))

        return cls.instances[code]

    @classmethod
    def values(cls):
        return cls.instances.values()

    @classmethod
    def db_type(cls):
        """
        :return: SQL Alchemy Type Decorator to pass into database column in model
        """
        class StatusTypeDecorator(sqlalchemy.types.TypeDecorator):
            impl = sqlalchemy.types.String

            def process_bind_param(self, value, dialect):
                if value is None:
                    return None

                if not isinstance(value, cls):
                    raise ValueError('Value must be a valid {}, {} given'
                                     .format(cls.__name__, type(value)))
                return value.code

            def process_result_value(self, value, dialect):
                if value is None:
                    return None

                return cls.from_code(value)

        return StatusTypeDecorator

